"""
Implement the derived class for MaNGA datacubes.

.. todo::

    - Further optimize calculation of transfer matrix/
    - Make DRP file class flexible to linear or log-linear wavelength
      sampling?  Incorporate into MODE?
    - Reconstructed broad-band images and PSFs are *not* restructured in
      the CUBE files!  
    - Image reconstruction has transpose sense wrt DRP output!
    - Add logging
    - Need to be clear about which functions use the RSS spectra to
      create CUBE related data, like the covariance matrix and
      instrumental dispersion calculations.
    - Computing the approximate covariance cube is currently not
      possible with only the CUBE on disk.  There's a logic problem that
      needs to be fixed.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import time
from pathlib import Path
import warnings

from IPython import embed

import numpy

from astropy.io import fits
from astropy.wcs import WCS

from ..config.manga import MaNGAConfig
from ..util.drpbitmask import DRPFitsBitMask, DRPQuality3DBitMask
from ..util.parser import DefaultConfig
from ..util.constants import DAPConstants
from ..util.covariance import Covariance
from ..spectra import MaNGARSS
from .datacube import DataCube


# NOTE: Because of inheriting from MaNGAConfig and the specific code in the
# MaNGAConfig.from_file method, beware of adding a `from_file` method to this!
# It shouldn't be a problem though, because the main instantiation method is
# effectively the "from_file" method.
class MaNGADataCube(MaNGAConfig, DataCube):
    r"""
    Container class for a MaNGA datacube.

    For additional description and attributes, see the two base
    classes.

    See :func:`from_plateifu` to instantiate using the plate and ifu
    numbers. See
    :func:`~mangadap.datacube.datacube.DataCube.from_config` to
    instantiate from a configuration file.

    Note that ``z``, ``vdisp``, ``ell``, ``pa``, and ``reff`` are
    saved to the :attr:`~mangadap.datacube.datacube.DataCube.meta`
    dictionary.

    Args:
        ifile (:obj:`str`, `Path`_):
            The file with the MaNGA datacube. The name is expected to
            follow the nominal naming convention, which is used to
            parse the plate, ifudesign, and whether or not the
            datacube has been binned logarithmically in wavelength.
        z (:obj:`float`, optional):
            Estimated bulk redshift. If None, some of the DAP
            analysis modules will fault.
        vdisp (:obj:`float`, optional):
            Estimated velocity dispersion. If None, some of the DAP
            analysis modules will assume an initial guess of 100
            km/s.
        ell (:obj:`float`, optional):
            Characteristic isophotal ellipticity (1-b/a). If None,
            some of the DAP modules will issue a warning and continue
            by assuming ``ell=0``.
        pa (:obj:`float`, optional):
            Characteristic isophotal position angle (through E from
            N). If None, some of the DAP modules will issue a warning
            and continue by assuming ``pa=0``.
        reff (:obj:`float`, optional):
            Effective (half-light) radius in arcsec. If None, some of
            the DAP modules will issue a warning and continue by
            assuming ``reff=1``.
        sres_ext (:obj:`str`, optional):
            The extension to use when constructing the spectral
            resolution vectors. See
            :func:`~mangadap.config.manga.MaNGAConfig.spectral_resolution`.
        sres_fill (:obj:`bool`, optional):
            Fill masked values by interpolation. Default is to leave
            masked pixels in returned array.
        covar_ext (:obj:`str`, optional):
            Extension to use as the single spatial correlation matrix
            for all wavelength channels, read from the DRP file. For
            generating the covariance matrix directly for an
            arbitrary wavelength channel using the RSS file, see
            :func:`mangadap.datacube.datacube.DataCube.covariance_matrix`.
    """
    def __init__(self, ifile, z=None, vdisp=None, ell=None, pa=None, reff=None, sres_ext=None,
                 sres_fill=True, covar_ext=None):

        # Use the configuration class method to construct the configuration
        # based on the input data file
        cfg = MaNGAConfig.from_file(ifile)
        # and then use the parent class init to set it to self.  NOTE: This is
        # round about, but the way I got it to work.  There must be a better way
        # to do this...
        MaNGAConfig.__init__(self, cfg.plate, cfg.ifudesign, mode=cfg.mode, log=cfg.log,
                             drpver=cfg.drpver, redux_path=cfg.redux_path,
                             chart_path=cfg.chart_path, directory_path=cfg.directory_path)

        # Collect the metadata into a dictionary
        meta = {}
        meta['z'] = z
        meta['vdisp'] = vdisp
        meta['ell'] = ell
        meta['pa'] = pa
        meta['reff'] = reff

        # Define the relevant BitMask object
        bitmask = DRPFitsBitMask(mode='CUBE')

        # Open the file and initialize the DataCube base class
        with fits.open(ifile) as hdu:
            # Read covariance first
            covar = None if covar_ext is None \
                        else Covariance.from_fits(hdu, ivar_ext=None, covar_ext=covar_ext,
                                                  impose_triu=True, correlation=True)

            print('Reading MaNGA datacube data ...', end='\r')
            prihdr = hdu[0].header
            fluxhdr = hdu['FLUX'].header
            # NOTE: Need to keep a log of what spectral resolution
            # vectors were used so that the same vectors are read from
            # the RSS file, if/when it is loaded.
            self.sres_ext, sres = MaNGAConfig.spectral_resolution(hdu, ext=sres_ext, fill=sres_fill)
            self.sres_fill = sres_fill
            sres = sres.filled(0.0)

            # Pull the Galactic E(B-V) from the header and add it to the metadata
            meta['ebv'] = prihdr['EBVGAL']
            # Determine if the datacube was flagged as having CRITICAL DRP failures
            meta['drpcrit'] = DRPQuality3DBitMask().flagged(prihdr['DRP3QUAL'], flag='CRITICAL')

            # NOTE: Transposes are done here because of how the data is
            # read from the fits file. The covariance is NOT transposed
            # accordingly because of how the correlation matrices are
            # stored by the DRP. Some care needs to be taken here
            # though because the transpose changes the array memory
            # storage from C to Fortran contiguous. However, everything
            # is expected to be okay because numpy array-flattening
            # always performs a C-like flattening, even if the memory
            # storage is Fortran contiguous.
            DataCube.__init__(self, hdu['FLUX'].data.T, wave=hdu['WAVE'].data,
                              ivar=hdu['IVAR'].data.T, mask=hdu['MASK'].data.T, bitmask=bitmask,
                              sres=sres.T, covar=covar, wcs=WCS(header=fluxhdr, fix=True),
                              pixelscale=0.5, log=self.log, meta=meta, prihdr=prihdr,
                              fluxhdr=fluxhdr, name=f'{self.plate}-{self.ifudesign}')
        print('Reading MaNGA datacube data ... DONE')

    @classmethod
    def from_plateifu(cls, plate, ifudesign, log=True, drpver=None, redux_path=None,
                      chart_path=None, directory_path=None, **kwargs):
        """
        Construct a :class:`mangadap.datacube.manga.MaNGADataCube`
        object based on its plate-ifu designation.

        This is a simple wrapper function that calls
        :func:`default_paths` to construct the file names, and then
        calls the class instantiation method.

        Args:
            plate (:obj:`int`):
                Plate number
            ifudesign (:obj:`int`):
                IFU design
            log (:obj:`bool`, optional):
                Use the datacube that is logarithmically binned in
                wavelength.
            drpver (:obj:`str`, optional):
                DRP version, which is used to define the default DRP
                redux path. Default is defined by
                :func:`mangadap.config.manga.drp_version`
            redux_path (:obj:`str`, optional):
                The path to the top level directory containing the
                DRP output files for a given DRP version. Default is
                defined by
                :func:`mangadap.config.manga.drp_redux_path`.
            chart_path (:obj:`str`, optional):
                The path to the directory containing the finding chart images
                for this plate.  Default is defined by
                :func:`mangadap.config.manga.drp_finding_chart_path`.
            directory_path (:obj:`str`, optional):
                The exact path to the DRP file. Default is defined by
                :func:`mangadap.config.manga.drp_directory_path`.
                Providing this ignores anything provided for
                ``drpver`` or ``redux_path``.
            **kwargs:
                Keyword arguments passed directly to the primary
                instantiation method; see
                :class:`mangadap.datacube.manga.MaNGADataCube` or
                :class:`mangadap.spectra.manga.MaNGARSS`.
        """
        # Use the input to instantiate the configuration
        cfg = MaNGAConfig(plate, ifudesign, log=log, drpver=drpver, redux_path=redux_path,
                          chart_path=chart_path, directory_path=directory_path)
        # Then use the configuration to instantiate the object
        return cls(str(cfg.file_path), **kwargs)

    @classmethod
    def from_config(cls, cfgfile):
        """
        Construct a :class:`mangadap.datacube.manga.MaNGADataCube` or
        :class:`mangadap.spectra.manga.MaNGARSS` object from a
        configuration file.

        Using the data read from the configuration file, the method
        instantiates the class using
        :func:`mangadap.spectra.manga.MaNGARSS.from_plateifu` or
        :func:`mangadap.datacube.manga.MaNGADataCube.from_plateifu`.
        This method will therefore fault for this base class!

        The format of the configuration file is:

        .. todo::

            Fill this in.

        Args:
            cfgfile (:obj:`str`):
                Configuration file
            drpver (:obj:`str`, optional):
                DRP version, which is used to define the default DRP
                redux path. Overrides any value in the configuration
                file.
            redux_path (:obj:`str`, optional):
                The path to the top level directory containing the
                DRP output files for a given DRP version. Overrides
                any value in the configuration file.
            directory_path (:obj:`str`, optional):
                The exact path to the DRP file. Providing this
                ignores anything provided for ``drpver`` or
                ``redux_path``. Overrides any value in the
                configuration file.
        """
        # First use the base class method to instantiate the configuration
        cfg = MaNGAConfig.from_config(cfgfile, mode='CUBE')

        # Then read the file (again...) to get the relevant keyword arguments
        full_cfg = DefaultConfig(cfgfile, interpolate=True)
        kwargs = {}
        kwargs['z'] = full_cfg.getfloat('z')
        kwargs['vdisp'] = full_cfg.getfloat('vdisp')
        kwargs['ell'] = full_cfg.getfloat('ell')
        kwargs['pa'] = full_cfg.getfloat('pa')
        kwargs['reff'] = full_cfg.getfloat('reff')
        kwargs['sres_ext'] = full_cfg.get('sres_ext')
        kwargs['sres_fill'] = full_cfg.getbool('sres_fill', default=True)
        kwargs['covar_ext'] = full_cfg.get('covar_ext')
        return cls(str(cfg.file_path), **kwargs)

    def load_rss(self, force=False):
        """
        Try to load the source row-stacked spectra for this datacube.

        If :attr:`~mangadap.datacube.datacube.DataCube.rss` is not
        None, this method does not attempt to reload it, unless
        ``force`` is True.

        The method first looks for the relevant file in the same
        directory with the datacube. If the file is not there, it
        uses the DRP version in the header of the datacube file and
        the paths defined by the keyword arguments to construct the
        nominal path to the RSS file. If the file is also not there,
        the method faults.

        Nothing is returned. If successful, the method initializes
        the row-stacked spectra object
        (:class:`mangadap.spectra.manga.MaNGARSS`) to
        :attr:`~mangadap.datacube.datacube.DataCube.rss`.

        Args:
            force (:obj:`bool`, optional):
                Reload the row-stacked spectra if
                :attr:`~mangadap.datacube.datacube.DataCube.rss` is
                not None.
        """
        if self.rss is not None and not force:
            return

        # Get the RSS counterpart
        rss_cfg = MaNGAConfig(self.plate, self.ifudesign, mode='RSS', log=self.log,
                              directory_path=self.directory_path)
        if not rss_cfg.file_path.exists():
            # Not in this directory.  Check the nominal directory
            warnings.warn(f'Could not find: {rss_file_path}.  Trying the default path.')
            rss_cfg = MaNGAConfig(self.plate, self.ifudesign, mode='RSS', log=self.log)
        if not rss_cfg.file_path.exists():
            raise FileNotFoundError(f'Could not find: {rss_file_path}')

        self.rss = MaNGARSS(str(rss_cfg.file_path), sres_ext=self.sres_ext,
                            sres_fill=self.sres_fill)

    def mean_sky_coordinates(self, center_coo=None, offset='OBJ'):
        """
        Calculate the sky coordinates for each spaxel.

        This is a simple wrapper for
        :func:`mangadap.datacube.datacube.DataCube.mean_sky_coordinates`
        that passes the relevant set of center coordinates; see that
        function for the default behavior. Any provided coordinates
        (see ``center_coo``) take precedence.

        Args:
            center_coo (:obj:`tuple`, optional):
                A two-tuple with the coordinates in right-ascension
                and declination for the coordinate-frame origin. If
                None, no offset is performed.
            offset (:obj:`str`, optional):
                The offset coordinates to use from the primary
                header. Must be either 'obj' or 'ifu'. If None, no
                header coordinates are used.

        Returns:
            :obj:`tuple`: Two `numpy.ndarray`_ objects with the RA
            and declination of each pixel in degrees, or its offset
            from the center in arcseconds. In both cases the shape of
            the returned arrays matches the spatial shape of the
            datacube.
        """
        if center_coo is not None and offset is not None:
            warnings.warn('Center coordinates provided directly.  Ignore other offset requests.')
        if offset is not None and offset.upper() not in ['OBJ', 'IFU']:
            raise ValueError('Offset must be None, obj or ifu.')
        _center_coo = (self.prihdr['{0}RA'.format(offset.upper())],
                       self.prihdr['{0}DEC'.format(offset.upper())]) \
                           if center_coo is None and offset is not None else center_coo
        return super().mean_sky_coordinates(center_coo=_center_coo)

    def approximate_correlation_matrix(self, sigma_rho=1.92, rlim=None, rho_tol=None, redo=False):
        """
        Construct an approximate correlation matrix for the datacube.

        This is a simple wrapper for
        :func:`mangadap.datacube.datacube.DataCube.approximate_correlation_matrix`
        that defaults to the expected values for a MaNGA datacube.
        See that method for a description of the arguments. The
        default ``sigma_rho`` is from Westfall et al. (2019, AJ, 158,
        231). The default ``rlim`` is provided by
        :func:`mangadap.config.manga.MaNGAConfig.regrid_rlim`.
        """
        _rlim = MaNGAConfig.regrid_rlim() if rlim is None else rlim
        return super().approximate_correlation_matrix(sigma_rho, _rlim, rho_tol=rho_tol, redo=redo)

    def approximate_covariance_matrix(self, channel, sigma_rho=1.92, rlim=None, rho_tol=None,
                                      csr=False, quiet=False):
        """
        Construct an approximate covariance matrix.
        
        This is a simple wrapper for
        :func:`mangadap.datacube.datacube.DataCube.approximate_covariance_matrix`
        that defaults to the expected values for a MaNGA datacube.
        See that method for a description of the arguments. The
        default ``sigma_rho`` is from Westfall et al. (2019, AJ, 158,
        231). The default ``rlim`` is provided by
        :func:`~mangadap.config.manga.MaNGAConfig.regrid_rlim`.
        """
        _rlim = MaNGAConfig.regrid_rlim() if rlim is None else rlim
        return super().approximate_covariance_matrix(channel, sigma_rho=sigma_rho, rlim=_rlim,
                                                     rho_tol=rho_tol, csr=csr, quiet=quiet)

    def approximate_covariance_cube(self, channels=None, sigma_rho=1.92, rlim=None, rho_tol=None,
                                    csr=False, quiet=False):
        """
        Construct approximate covariance matrices for multiple channels.
        
        This is a simple wrapper for
        :func:`mangadap.datacube.datacube.DataCube.approximate_covariance_cube`
        that defaults to the expected values for a MaNGA datacube.
        See that method for a description of the arguments. The
        default ``sigma_rho`` is from Westfall et al. (2019, AJ, 158,
        231). The default ``rlim`` is provided by
        :func:`mangadap.config.manga.MaNGAConfig.regrid_rlim`.
        """
        _rlim = MaNGAConfig.regrid_rlim() if rlim is None else rlim
        return super().approximate_covariance_cube(channels=channels, sigma_rho=sigma_rho,
                                                   rlim=_rlim, rho_tol=rho_tol, csr=csr,
                                                   quiet=quiet)




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""

Provides a set of parsing utility functions.

.. todo::
    - Add function that will parse the default MaNGA fits file name (see
      :func:`mangadap.config.defaults.manga_fits_root`).

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
from pathlib import Path
from os import environ
import warnings
from IPython import embed
from configparser import ConfigParser, ExtendedInterpolation

def arginp_to_list(inp, evaluate=False, quiet=True):
    """
    Separate a list of comma-separated values in the input string to
    a :obj:`list`.

    Args:
        inp (:obj:`str`, :obj:`list`):
            Input string with a list of comma-separated values
        evaluate (:obj:`bool`, optional):
            Attempt to evaluate the elements in the list using
            `eval`_.
        quiet (:obj:`bool`, optional):
            Suppress terminal output.

    Returns:
        :obj:`list`: The list of the comma-separated values
    """
    out = inp

    # Simply return a None value
    if out is None:
        return out

    # If the input value is a string, convert it to a list of strings
    if isinstance(out, str):
        out = out.replace("[", "")
        out = out.replace("]", "")
        out = out.replace(" ", "")
        out = out.split(',')

    # If the input is still not a list, make it one
    if not isinstance(out, list):
        out = [out]

    # Evaluate each string, if requested
    if evaluate:
        n = len(out)
        for i in range(0,n):
            try:
                tmp = eval(out[i])
            except (NameError, TypeError) as e:
                if not quiet:
                    warnings.warn(f'Could not evaluate value {out[i]}. Skipping.')
            else:
                out[i] = tmp

    return out


# TODO: Replace with ', '.join([str(f) for f in flist])
def list_to_csl_string(flist):
    """
    Convert a list to a comma-separated string; i.e. perform the inverse
    of :func:`arginp_to_list`.

    Args:
        flist (:obj:`list`):
            List to convert to a comma-separated string

    Returns:
        :obj:`str`: String with the values in the input list
        converted to strings separated by commas
    """
    n = len(flist)
    out=str(flist[0])
    for i in range(1,n):
        out += ', '+str(flist[i])
    return out


def parse_drp_file_name(name):
    """
    Parse the name of a DRP file to provide the plate, ifudesign, and
    mode.

    Args:
        name (str): Name of the DRP file.

    Returns:
        int, int, str: The plate, ifudesign, and mode ('RSS' or 'CUBE')
        of the DRP file pulled from the file name.

    Raises:
        TypeError: Raised if *name* is not a string.
        ValueError: Raised if if the file name does not look like a DRP
            file because it does not include 'manga-', '-LOG', or
            '.fits.gz'.
    """

    if (type(name) != str):
        raise TypeError("File name must be a string.")
    if (str.find(name, 'manga-') == -1 or str.find(name, '-LOG') == -1 or 
        str.find(name, '.fits.gz') == -1):
        raise ValueError("String does not look like a DRP fits file name.")

    plate_start = str.find(name, '-')+1
    plate_end = str.find(name,'-',plate_start)
    plate = long(name[plate_start:plate_end])

    ifudesign_end = str.find(name,'-',plate_end+1)
    ifudesign = long(name[plate_end+1:ifudesign_end])

    mode_start = str.find(name,'LOG')+3
    mode = name[mode_start:str.find(name,'.fits.gz')]

    return plate, ifudesign, mode


#def parse_dap_file_name(name):
#    """
#    Parse the name of a DAP file and return the plate, ifudesign, mode,
#    binning type, and iteration number.
#
#    Args:
#        name (str): Name of the DAP file.
#
#    Returns:
#        int, int, str, str, int : The plate, ifudesign, mode ('RSS' or
#        'CUBE'), bin type, and iteration number of the DAP file, pulled
#        from the name of the file.
#
#    Raises:
#        TypeError: Raised if *name* is not a string.
#        ValueError: Raised if if the file name does not look like a DRP
#            file because it does not include 'manga-', '-BIN', or
#            '.fits'.
#    """
#    if (type(name) != str):
#        raise TypeError("File name must be a string.")
#    if (str.find(name, 'manga-') == -1 or str.find(name, '-LOG') == -1
#        or str.find(name, 'BIN-') == -1 or str.find(name, '.fits') == -1):
#        raise ValueError("String does not look like a DAP fits file name.")
#
#    plate_start = str.find(name, '-')+1
#    plate_end = str.find(name,'-',plate_start)
#    plate = int(name[plate_start:plate_end])
#
#    ifudesign_end = str.find(name,'-',plate_end+1)
#    ifudesign = int(name[plate_end+1:ifudesign_end])
#
#    mode_start = str.find(name,'LOG')+3
#    mode_end = str.find(name,'_BIN-')
#    mode = name[mode_start:mode_end]
#
#    bintype_end = str.find(name,'-',mode_end+5)
#    bintype = name[mode_end+5:bintype_end]
#
#    niter_end = str.find(name,'.fits',bintype_end)
#    niter = int(name[bintype_end+1:niter_end])
#
#    return plate, ifudesign, mode, bintype, niter

    
class DefaultConfig:
    """
    A wrapper for the ConfigParser class that handles None values and
    provides some convenience functions.
    """
    def __init__(self, f=None, interpolate=False):
        # TODO: May not need extended interpolation anymore...
        self.cnfg = ConfigParser(environ, allow_no_value=True,
                                 interpolation=ExtendedInterpolation()) \
                        if interpolate else ConfigParser(allow_no_value=True)
        if f is None:
            return
        self.read(f)

    def __getitem__(self, key):
        return self.cnfg['default'][key]

    def __iter__(self):
        return self.cnfg.options('default').__iter__()

    def read(self, f):
        _f = Path(f).resolve()
        if not _f.exists():
            raise FileNotFoundError(f'Configuration file not found: {_f}')
        self.cnfg.read(f)

    def keyword_specified(self, key):
        return key in self and self[key] is not None

    def all_required(self, keys):
        for k in keys:
            if k not in self or self[k] is None:
                return False
        return True

    def get(self, key, default=None):
        return default if key not in self or self.cnfg['default'][key] is None \
                        else self.cnfg['default'][key]

    def getint(self, key, default=None):
        return default if key not in self or self.cnfg['default'][key] is None \
                        else self.cnfg['default'].getint(key)

    def getbool(self, key, default=None):
        return default if key not in self or self.cnfg['default'][key] is None \
                        else self.cnfg['default'].getboolean(key)

    def getfloat(self, key, default=None):
        return default if key not in self or self.cnfg['default'][key] is None \
                        else self.cnfg['default'].getfloat(key)

    def getlist(self, key, evaluate=False, default=None):
        if key not in self or self.cnfg['default'][key] is None:
            return default
        return [ eval(e.strip()) if evaluate else e.strip() \
                    for e in self.cnfg['default'][key].split(',') ]







from ..util.pkg import all_subclasses
from . import scriptbase

# The import of all the script modules here is what enables the dynamic
# compiling of all the available scripts below
from . import calculate_covariance
from . import construct_dapall
from . import dap_status
from . import dapall_qa
from . import find_repeat_observations
from . import fit_residuals
from . import manga_dap_inspector
from . import manga_dap
from . import plate_fit_qa
from . import ppxffit_qa
#from . import rundap
from . import spotcheck_dap_maps
from . import template_flux_norm
from . import write_dap_config

# Build the list of script classes
def script_classes():
    import numpy as np

    # Recursively collect all subclasses
    scr_c = np.array(list(all_subclasses(scriptbase.ScriptBase)))
    scr_n = np.array([c.name() for c in scr_c])
    # Construct a dictionary with the script name and class
    srt = np.argsort(scr_n)
    return dict([ (n,c) for n,c in zip(scr_n[srt],scr_c[srt])])

mangadap_scripts = list(script_classes().keys())




import os
from IPython import embed

from mangadap.proc import util
from mangadap.config import defaults


def test_database_key():
    assert util.get_database_key('junk') == 'JUNK'
    assert util.get_database_key('test.par') == 'TEST'
    assert util.get_database_key('/path/to/test.par') == 'TEST' 


def test_select_database():
    d = defaults.dap_data_root() / 'emission_lines'
    assert os.path.split(util.select_database('ELPMPL11', d))[1] == 'elpmpl11.par'





from IPython import embed

from mangadap.par.bandheadindexdb import BandheadIndexDB


def test_read():
    dbs = BandheadIndexDB.available_databases()
    for key in dbs.keys():
        bhddb = BandheadIndexDB.from_key(key)


def test_basic():
    bhddb = BandheadIndexDB.from_key('BHBASIC')
    assert len(bhddb) == 3, 'Incorrect number of bandhead indices'
    assert 'Dn4000' in bhddb['name'], 'Does not contain Dn4000 in list'





#!/usr/bin/env python3

import os
import time
import numpy

from matplotlib import pyplot, ticker, rc, colors

from astropy.io import fits
import astropy.constants

from mangadap.proc.util import growth_lim
from mangadap.util.bitmask import BitMask

#-----------------------------------------------------------------------------
def init_ax(fig, pos):
    ax = fig.add_axes(pos, facecolor='0.98')
    ax.grid(True, which='major', color='0.9', zorder=0, linestyle='-')
    ax.minorticks_on()
    ax.tick_params(which='major', length=6) #, direction='in')
    ax.tick_params(which='minor', length=3) #, direction='in')
    return ax

def mgfe_hbeta(drpall, dapall):
    indx = (dapall['DAPDONE'] == 1) & (dapall['DAPTYPE'] == 'HYB10-GAU-MILESHC')
    indx &= (drpall['nsa_sersic_mass'][dapall['DRPALLINDX']] > 0)

    indx &= (dapall['SPECINDEX_1RE'][:,12] > -999)
    indx &= (dapall['SPECINDEX_1RE'][:,13] > -999)
    indx &= (dapall['SPECINDEX_1RE'][:,14] > -999)
    indx &= (dapall['SPECINDEX_1RE'][:,8] > -999)

    mgfe = numpy.ma.sqrt(dapall['SPECINDEX_1RE'][indx,12] * (0.72*dapall['SPECINDEX_1RE'][indx,13] 
                        + 0.28*dapall['SPECINDEX_1RE'][indx,13]))

    mass_lim = [1e9, 7e11]
    mgfe_lim = [0, 4.5]
    hb_lim = [0.5, 5.9]

    font = { 'size' : 16 }
    rc('font', **font)

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    ax = init_ax(fig, [0.2, 0.2, 0.6, 0.6])
    ax.set_xlim(mgfe_lim)
    ax.set_ylim(hb_lim)
    ax.tick_params(axis='x', which='both', direction='in')

    sc = ax.scatter(mgfe, dapall['SPECINDEX_1RE'][indx,8],
                    marker='.', s=30, lw=0, zorder=2,
                    c=drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][indx]],
                    norm=colors.LogNorm(vmin=mass_lim[0], vmax=mass_lim[1]),
                    cmap='viridis')

    ax.text(-0.15, 0.5, r'$\langle {\rm H}\beta\rangle_e$ [$\AA$, in absorption]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
            rotation='vertical')

    ax.text(0.5, -0.12, r'[$\langle$Mgb$\rangle_e\cdot$(0.72 $\langle$Fe5270$\rangle_e$ + 0.28 $\langle$Fe5335$\rangle_e$)]$^{1/2}$',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes)
    ax.text(0.5, -0.19, r'[$\AA$, in absorption]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes)

    cax = fig.add_axes([0.58, 0.71, 0.20, 0.01])
    cb = pyplot.colorbar(sc, cax=cax, orientation='horizontal', ticks=[1e9,1e10,1e11])
    cax.text(0.5, 3.0, r'$\mathcal{M}_\ast$ [$h^{-2} \mathcal{M}_\odot$]',
             ha='center', va='center', transform=cax.transAxes) #, fontsize=10)

    ofile = '../ms/figs/mgfe_hbeta.pdf'
    fig.canvas.print_figure(ofile, bbox_inches='tight')
    fig.clear()
    pyplot.close(fig)

#-----------------------------------------------------------------------------

if __name__ == '__main__':
    t = time.clock()

    dapall_file = '/Users/westfall/Work/MaNGA/testing/dr15/dapall-v2_4_3-2.2.1.fits'
    dapall = fits.open(dapall_file)['DAPALL'].data

    drpall_file = '/Users/westfall/Work/MaNGA/testing/dr15/drpall-v2_4_3.fits'
    drpall = fits.open(drpall_file)[1].data

    mgfe_hbeta(drpall, dapall)

    print('Elapsed time: {0} seconds'.format(time.clock() - t))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Container class for a database of emission-line parameters, as well as
support classes and functions.

Class usage examples
--------------------

To define an emission line::

    from mangadap.par.emissionlinedb import EmissionLinePar
    p = EmissionLinePar(index=44, name='Ha', restwave=6564.632, action='f', line='l',
                        flux=1.0, vel=0.0, sig=10., mode='f')

More often, however, you will want to define an emission-line
database using an SDSS parameter file. To do so, you can use one of
the default set of available emission-line databases::

    from mangadap.par.emissionlinedb import EmissionLineDB
    print(EmissionLineDB.available_databases())
    emldb = EmissionLineDB.from_key('ELPMPL9')

The above call uses the
:func:`~mangadap.par.spectralfeaturedb.SpectralFeatureDB.from_key`
method to define the database using its keyword and the database
provided with the MaNGA DAP source distribution. You can also define
the database directly for an SDSS-style parameter file::

    from mangadap.par.emissionlinedb import EmissionLineDB
    emldb = EmissionLineDB('/path/to/emission/line/database/myeml.par')

The above will read the file and set the database keyword to
'MYEML' (i.e., the capitalized root name of the ``*.par`` file).
See :ref:`emissionlines` for the format of the parameter file.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""

import os
import glob

from IPython import embed

import numpy

from pydl.goddard.astro import airtovac
from pydl.pydlutils.yanny import yanny

from .parset import KeywordParSet
from .spectralfeaturedb import SpectralFeatureDB
from ..proc import util
from ..util.datatable import DataTable


class EmissionLinePar(KeywordParSet):
    r"""
    Metadata used to define an emission line for the DAP to fit
    parametrically.

    These parameters are only for a single emission line; see
    :class:`~mangadap.par.emissionlinedb.EmissionLineDB` for setting a full
    line list of parameters.

    .. todo::
        - Point to/add a description the paramters.

    """
    def __init__(self, index=None, name=None, restwave=None, action=None, tie_index=None,
                 tie_par=None, blueside=None, redside=None):
        
        in_fl = [ int, float ]
        action_options = [ 'i', 'f', 'm', 's']
        arr_like = [ numpy.ndarray, list ]

        _name = None if name is None else name.strip()
        _action = None if action is None else action.strip()
        
        pars =     ['index', 'name', 'restwave', 'action', 'tie_index', 'tie_par', 'blueside',
                    'redside']
        values =   [index, _name, restwave, _action, tie_index, tie_par, blueside, redside]
        defaults = [None, None, None, 'f', None, None, None, None]
        options =  [None, None, None, action_options, None, None, None, None]
        dtypes =   [int, str, in_fl, str, int, arr_like, arr_like, arr_like]

        descr = ['An index used to refer to the line for tying parameters; must be >0.',
                 'A name for the line.',
                 'The rest wavelength of the line in angstroms *in vacuum*.',
                 'Describes how the line should be treated.  See ' \
                    ':ref:`emission-line-modeling-action`. Default is ``f``.',
                 'Index of the line to which parameters for this line are tied.  ' \
                    'See :ref:`emission-line-modeling-tying`.',
                 'Details of how each model parameter for this line is tied to the reference ' \
                    'line.  See :ref:`emission-line-modeling-tying`.',
                 'A two-element vector with the starting and ending wavelength for a bandpass ' \
                    'blueward of the emission line, which is used to set the continuum level ' \
                    'near the emission line when calculating the equivalent width.',
                 'A two-element vector with the starting and ending wavelength for a bandpass ' \
                    'redward of the emission line, which is used to set the continuum level ' \
                    'near the emission line when calculating the equivalent width.']

        super().__init__(pars, values=values, defaults=defaults, options=options, dtypes=dtypes,
                         descr=descr)

        self._check()

    def _check(self):
        if self.data['index'] is not None and self.data['index'] <= 0:
            raise ValueError('Index must be larger than 0.')


class EmissionLineDefinitionTable(DataTable):
    """
    A :class:`~mangadap.util.datatable.DataTable` with the data from an
    :class:`~mangadap.par.emissionlinedb.EmissionLineDB` object.

    Table includes:

    .. include:: ../tables/emissionlinedefinitiontable.rst
    
    Args:
        name_len (:obj:`int`, optional):
            The maximum length of any of the emission line names.
        tie_len (:obj:`int`, optional):
            The maximum length of a tying string. Not sure if this is needed.
        shape (:obj:`int`, :obj:`tuple`, optional):
            The shape of the initial array. If None, the data array
            will not be instantiated; use
            :func:`~mangadap.util.datatable.DataTable.init` to
            initialize the data array after instantiation.
    """
    def __init__(self, name_len=1, tie_len=1, shape=None):
        # NOTE: This requires python 3.7 to make sure that this is an "ordered"
        # dictionary.
        datamodel = dict(ID=dict(typ=int, shape=None, descr='Emission line ID number'),
                         NAME=dict(typ='<U{0:d}'.format(name_len), shape=None,
                                   descr='Name of the emission line'),
                         RESTWAVE=dict(typ=float, shape=None,
                                       descr='Rest wavelength of the emission line'),
                         ACTION=dict(typ='<U1', shape=None,
                                     descr='Action to take for this emission line; see '
                                           ':ref:`emission-line-modeling-action`'),
                         TIE_ID=dict(typ=int, shape=None,
                                     descr='ID of the line to which this one is tied.'),
                         TIE_FLUX=dict(typ='<U{0:d}'.format(tie_len), shape=None,
                                      descr='Tying parameter for the flux of each line.'),
                         TIE_VEL=dict(typ='<U{0:d}'.format(tie_len), shape=None,
                                      descr='Tying parameter for the velocity of each line.'),
                         TIE_SIG=dict(typ='<U{0:d}'.format(tie_len), shape=None,
                                      descr='Tying parameter for the dispersion of each line.'))

        keys = list(datamodel.keys())
        super().__init__(keys, [datamodel[k]['typ'] for k in keys],
                         element_shapes=[datamodel[k]['shape'] for k in keys],
                         descr=[datamodel[k]['descr'] for k in keys], shape=shape)


class EmissionLineDB(SpectralFeatureDB):
    r"""
    Basic container class for the database of emission-line parameters.

    Each row of the database is parsed using
    :class:`~mangadap.par.emissionlinedb.EmissionLinePar`. For the format of
    the input file, see :ref:`emissionlines-modeling`.

    The primary instantiation requires the SDSS parameter file with the
    emission-line data. To instantiate using a keyword, use the
    :func:`~mangadap.par.spectralfeaturedb.SpectralFeatureDB.from_key` method
    of the base class; see the base class for instantiation and additional
    attributes.

    Attributes:
        key (:obj:`str`):
            Database signifying keyword
        file (:obj:`str`):
            File with the emission-line data
        size (:obj:`int`):
            Number of emission lines in the database. 
    """

    default_data_dir = 'emission_lines'
    """
    Name of the directory in the mangadap data path with the distributed
    emission-line databases.
    """

    def _parse_yanny(self):
        """
        Parse the yanny file (provided by :attr:`file`) for the emission-line
        database.

        Returns:
            :obj:`list`: The list of :class:`~mangadap.par.parset.ParSet`
            instances for each line of the database.
        """
        # Read the yanny file
        par = yanny(filename=self.file, raw=True)
        if len(par['DAPEML']['index']) == 0:
            raise ValueError('Could not find DAPEML entries in {0}!'.format(self.file))

        # Setup the array of emission line database parameters
        self.size = len(par['DAPEML']['index'])
        parlist = []
        for i in range(self.size):
            invac = par['DAPEML']['waveref'][i] == 'vac'

            tie_index = -1 if par['DAPEML']['tie'][i][0] == 'None' \
                            else int(par['DAPEML']['tie'][i][0])
            tie_par = [None if t == 'None' else t for t in par['DAPEML']['tie'][i][1:]]

            parlist += [EmissionLinePar(index=par['DAPEML']['index'][i],
                                       name=par['DAPEML']['name'][i],
                                       restwave=par['DAPEML']['restwave'][i] if invac 
                                                else airtovac(par['DAPEML']['restwave'][i]),
                                       action=par['DAPEML']['action'][i], tie_index=tie_index,
                                       tie_par=tie_par,
                                       blueside=par['DAPEML']['blueside'][i] if invac else \
                                            airtovac(numpy.array(par['DAPEML']['blueside'][i])),
                                       redside=par['DAPEML']['redside'][i] if invac else \
                                            airtovac(numpy.array(par['DAPEML']['redside'][i])))]
        return parlist

    def channel_names(self, dicttype=True):
        """
        Return a dictionary with the channel names as the dictionary
        key and the channel number as the dictionary value. If
        ``dicttype`` is False, a list is returned with just the
        string keys.
        """
        channels = ['{0}-{1}'.format(self.data['name'][i], int(self.data['restwave'][i])) 
                        for i in range(self.size)]
        return {n:i for i,n in enumerate(channels)} if dicttype else channels

    def to_datatable(self, quiet=False):
        """
        Constuct an
        :class:`~mangadap.par.emissionlinedb.EmissionLineDefinitionTable`
        instance with the line database.

        Args:
            quiet (:obj:`bool`, optional):
                Suppress terminal output

        Returns:
            :class:`~mangadap.par.emissionlinedb.EmissionLineDefinitionTable`:
            Table with the emission-line metadata.
        """
        name_len = numpy.amax([len(n) for n in self.data['name']])
        tie_len = numpy.amax([0 if t is None else len(t) 
                                for t in numpy.ravel(self.data['tie_par'])])

        # Instatiate the table data that will be saved defining the set
        # of emission-line moments measured
        db = EmissionLineDefinitionTable(name_len=name_len, tie_len=tie_len, shape=self.size)

        hk = ['ID', 'NAME', 'RESTWAVE', 'ACTION', 'TIE_ID']
        mk = ['index', 'name', 'restwave', 'action', 'tie_index']
        for _hk, _mk in zip(hk,mk):
            db[_hk] = self.data[_mk]

        db['TIE_FLUX'] = numpy.array([str(d) for d in self.data['tie_par'][:,0]])
        db['TIE_VEL'] = numpy.array([str(d) for d in self.data['tie_par'][:,1]])
        db['TIE_SIG'] = numpy.array([str(d) for d in self.data['tie_par'][:,2]])

        return db

    def tie_index_match(self):
        """
        Return a vector with the row indices for the tied parameters, which
        is more useful than the ID number of the tied line.
        """
        tie_indx = numpy.full(self.size, -1, dtype=int)
        indx = self.data['tie_index'] > 0
        tie_indx[indx] = [numpy.where(self.data['index'] == i)[0][0]
                            for i in self.data['tie_index'][indx]]
        return tie_indx

    @property
    def neml(self):
        """
        Number of emission lines in the database.
        """
        return self.nsets





# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
r"""
Provides a set of functions that define and return defaults used by the
MaNGA DAP, such as paths and file names.

----        

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""

from pathlib import Path
from pkg_resources import resource_filename


def dap_source_dir():
    from pkg_resources import resource_filename
    return Path(resource_filename('mangadap', '')).resolve()


def dap_data_root():
    """Return the root directory with the DAP data."""
    return Path(resource_filename('mangadap', 'data')).resolve()


def dap_config_root():
    """Return the root directory with the DAP config data."""
    return Path(resource_filename('mangadap', 'config')).resolve()


def sdss_maskbits_file():
    """Return the path to the sdss maskbits yanny file."""
    maskbits_file = dap_data_root() / 'sdss' / 'sdssMaskbits.par'
    return maskbits_file if maskbits_file.exists() else None


# USED BY:
#   - scripts/dapall_qa.py
#   - scripts/find_repeat_observations.py
#   - survey/dapall.py
def dapall_file(drpver=None, dapver=None, analysis_path=None):
    """
    Return the path to the DAPall file.

    Args:
        drpver (:obj:`str`, optional):
            DRP version.  Default is to use :func:`drp_version`.
        dapver (:obj:`str`, optional):
            DAP version.  Default is to use :func:`dap_version`.
        analysis_path (:obj:`str`, optional):
            Path to the root analysis directory.  Default is to use
            :func:`dap_analysis_path`

    Returns:
        :obj:`str`: Full path to the DAPall fits file.
    """
    _drpver = drp_version() if drpver is None else drpver
    _dapver = dap_version() if dapver is None else dapver
    _analysis_path = dap_analysis_path(drpver=_drpver, dapver=_dapver) \
                        if analysis_path is None else Path(analysis_path).resolve()
    return _analysis_path / f'dapall-{_drpver}-{_dapver}.fits'


# USED BY:
#   - scripts/rundap.py
def dap_file_root(plate, ifudesign, mode=None):
    """
    Generate the root name of the MaNGA DAP parameter and script
    files for a given plate/ifudesign/mode.
    
    Args:
        plate (:obj:`int`):
            Plate number
        ifudesign (:obj:`int`):
            IFU design number
        mode (:obj:`str`, optional):
            Mode of the DRP reduction; either ``RSS`` or ``CUBE``. If
            None, the mode is excluded from the file root.

    Returns:
        :obj:`str`: Root name for the DAP file:
        ``mangadap-[PLATE]-[IFUDESIGN]`` or
        ``mangadap-[PLATE]-[IFUDESIGN]-LOG[MODE]``
    """
    return f'mangadap-{plate}-{ifudesign}' if mode is None else \
                f'mangadap-{plate}-{ifudesign}-LOG{mode}'

    
# USED BY:
#   - scripts/rundap.py
def dap_config(plate, ifudesign, drpver=None, dapver=None, analysis_path=None,
               directory_path=None):
    """
    Return the full path to the DAP configuration file.

    The configuration file provides the input data necessary to
    instantiate a :class:`mangadap.datacube.manga.MaNGADataCube`.
    
    Args:
        plate (:obj:`int`):
            Plate number
        ifudesign (:obj:`int`):
            IFU design number
        drpver (:obj:`str`, optional):
            DRP version. Default is to use :func:`drp_version`.
        dapver (:obj:`str`, optional):
            DAP version. Default is to use :func:`dap_version`.
        analysis_path (:obj:`str`, optional): 
            Path to the root analysis directory. Default is to use
            :func:`dap_analysis_path`.
        directory_path (:obj:`str`, optional):
            Path to the directory with the DAP output files. Default
            is to use :func:`dap_common_path`

    Returns:
        :obj:`str`: Full path to the DAP par file
    """
    # Make sure the directory path is defined
    _directory_path = dap_common_path(plate=plate, ifudesign=ifudesign, drpver=drpver,
                                      dapver=dapver, analysis_path=analysis_path) \
                            if directory_path is None else Path(directory_path).resolve()
    # Set the name of the par file; put this in its own function?
    return _directory_path /  f'{dap_file_root(plate, ifudesign, "CUBE")}.ini'

    




from IPython import embed

import numpy
from astropy.io import fits

from mangadap.util.datatable import DataTable
from mangadap.proc.spectralfitting import StellarKinematicsFitDataTable
from mangadap.proc.spectralfitting import EmissionLineFitDataTable


def test_empty():
    keys = ['TEST', 'THIS']
    types = [ int, float]
    descr = ['This is a test', 'And so is this']
    db = DataTable(keys, types, descr=descr)
    assert db.data is None, 'No data should be defined'


def test_init():
    keys = ['TEST', 'THIS']
    types = [ int, float]
    descr = ['This is a test', 'And so is this']

    # Test at instantiation
    db = DataTable(keys, types, descr=descr, shape=10)
    assert db.data is not None, 'Data should have been instantiated'
    assert db.shape == (10,), 'Shape is incorrect'

    # Test after the fact
    db = DataTable(keys, types, descr=descr)
    db.init(10)
    assert db.data is not None, 'Data should have been instantiated'
    assert db.shape == (10,), 'Shape is incorrect'


def test_dtype():
    keys = ['TEST', 'THIS']
    types = [ int, float]
    descr = ['This is a test', 'And so is this']
    db = DataTable(keys, types, descr=descr, shape=10)

    _dtype = numpy.dtype([(k,t) for k,t in zip(keys,types)])
    assert db.dtype == _dtype, 'dtype is wrong'


def test_access():
    keys = ['TEST', 'THIS']
    types = [ int, float]
    descr = ['This is a test', 'And so is this']
    db = DataTable(keys, types, descr=descr, shape=10)

    assert numpy.array_equal(db['TEST'], numpy.zeros(10, dtype=int)), \
                'Instantiation array is wrong.'

    db['THIS'] = numpy.arange(10)
    assert numpy.array_equal(db['THIS'], numpy.arange(10, dtype=float)), \
                'Assignment went wrong.'

    db['THIS'][5] = 20.
    assert db['THIS'][5] == 20., 'Assignment of single value went wrong.'

    db[5]['THIS'] = 100.
    assert db['THIS'][5] == 100., 'Assignment of single value went wrong.'


def test_tohdu():
    keys = ['TEST', 'THIS']
    types = [ int, float]
    descr = ['This is a test', 'And so is this']
    db = DataTable(keys, types, descr=descr, shape=10)

    hdu = fits.HDUList([fits.PrimaryHDU(), db.to_hdu(name='TABLE')])

    assert isinstance(hdu['TABLE'], fits.BinTableHDU), 'Bad return type'
    assert numpy.array_equal(hdu['TABLE'].data['TEST'], db['TEST']), 'Arrays are different.'
    assert numpy.array_equal(hdu['TABLE'].data['THIS'], db['THIS']), 'Arrays are different.'


def test_torst():
    keys = ['TEST', 'THIS']
    types = [ int, float]
    descr = ['This is a test', 'And so is this']
    db = DataTable(keys, types, descr=descr, shape=10)
    # Just test for success of the method
    lines = db.to_rst_table()


def test_stellarkin_dt():
    db = StellarKinematicsFitDataTable(ntpl=20, nadd=8, nmult=8, nkin=2, mask_dtype=numpy.int16,
                                       shape=10)
    assert db.shape == (10,), 'Shape is wrong.'
    assert db['TPLWGT'].shape == (10,20), 'Number of template weights is wrong.'
    assert db['ADDCOEF'].shape == (10,8), 'Number of additive polynomial coefficients is wrong'
    assert db['KIN'][0].shape == (2,), 'Number of kinematic moments is wrong.'


def test_emission_dt():
    db = EmissionLineFitDataTable(20, 2, numpy.int16, shape=10)
    assert db.shape == (10,), 'Shape is wrong.'
    assert db.size == 10, 'Size is wrong'
    assert db['BINID'].shape == (10,), 'Number of bin IDs is wrong.'
    assert db['MASK'][0].shape == (20,), 'Number of maskbits is wrong'
    assert db['KIN'].shape == (10,20,2), 'Number of kinematic moments is wrong.'





#!/usr/bin/env python3

# Useful stuff:
#-----------------------------------------------------------------------
# ASTROPY CONSTANTS
# from astropy import constants
# constants.c.to('km/s').value
#-----------------------------------------------------------------------
# MATPLOTLIB TICKMARKS: ORDERING WILL MATTER TO OUTPUT
# from matplotlib.ticker import NullFormatter
# ax.xaxis.set_major_formatter(NullFormatter())
# ax.minorticks_on()
# ax.tick_params(which='major', length=6)
# ax.tick_params(which='minor', length=3)
#
# from matplotlib.ticker import MultipleLocator, FormatStrFormatter
# majorLocator   = MultipleLocator(20)
# majorFormatter = FormatStrFormatter('%d')
# minorLocator   = MultipleLocator(5)
# ax.xaxis.set_major_locator(majorLocator)
# ax.xaxis.set_major_formatter(majorFormatter)
# ax.xaxis.set_minor_locator(minorLocator)
#-----------------------------------------------------------------------
# MATPLOTLIB PRINT TO A FILE
# fig = pyplot.figure(1)
# fig.canvas.print_figure('out.pdf')
# fig.canvas.print_figure('test.pdf', bbox_inches='tight')
#-----------------------------------------------------------------------
# MATPLOTLIB LOG SCALE
# pyplot.xscale('log')
#-----------------------------------------------------------------------
# MATPLOTLIB TEXT IN PLOT VIEWPORT COORDINATES
# ax.text(-0.25, 0.50, r'$\Delta V$ (MIUSCAT-MILES)', horizontalalignment='center',
#         verticalalignment='center', transform=ax.transAxes, rotation='vertical')
#-----------------------------------------------------------------------
# MATPLOTLIB SET FIGURE ASPECT RATIO
# w,h = pyplot.figaspect(1)
# fig = pyplot.figure(figsize=(1.5*w,1.5*h))
#-----------------------------------------------------------------------
# MATPLOTLIB CREATE AXIS WITH VIEWPORT COORDINATES
# ax = pyplot.axes([ left, bottom+i*height, width, height ])
#   OR IN A FIGURE (with a gray background):
# ax = fig.add_axes([0.07, 0.66, 0.3, 0.3], axisbg='0.95')
#-----------------------------------------------------------------------
# MATPLOTLIB ADD GRID TO PANEL
# ax.grid(True, which='major', color='0.8', zorder=0, linestyle='-')
# ax.grid(True, which='minor', color='0.85', zorder=0, linestyle=':')
#-----------------------------------------------------------------------
# MATPLOTLIB CREATE PLOT PANEL GRID
# from matplotlib import gridspec
# gs = gridspec.GridSpec(100,100)
# ax = pyplot.subplot(gs[0:30,4:34])
#-----------------------------------------------------------------------
# MATPLOTLIB CHANGE FONT (OR OTHER PROPERTIES)
# from matplotlib import rc
# font = { 'size' : 8 }
# rc('font', **font)
#-----------------------------------------------------------------------
# MATPLOTLIB CREATE SCATTER PLOT WITH COLOR BAR
# import matplotlib.pyplot as plt
# cm = plt.cm.get_cmap('RdYlBu')
# xy = range(20)
# z = xy
# sc = plt.scatter(xy, xy, c=z, vmin=0, vmax=20, s=35, cmap=cm)
# plt.colorbar(sc)
# plt.show()
#-----------------------------------------------------------------------
# MATPLOTLIB SHADED BOX
# box_sig_x = [ sig_lim[0], sigi_x_1901, sigi_x_1901, sig_lim[0] ]
# box_sig_y = [ sig_lim[0], sig_lim[0], sigi_y_1901, sigi_y_1901 ]
# ax.fill_between( box_sig_x, box_sig_y, color='blue', alpha=0.1, zorder=1)
#-----------------------------------------------------------------------
# MATPLOTLIB OPPOSITE AXIS LABELS, LIMITS
# axt = ax.twinx()
# axt.set_xlim(alim)
# axt.set_ylim([1,45])
# axt.minorticks_on()
# axt.tick_params(which='major', length=10)
# axt.tick_params(which='minor', length=5)
# axt.xaxis.set_major_locator(MultipleLocator(0.05))
# axt.xaxis.set_minor_locator(MultipleLocator(0.01))
# axt.yaxis.set_major_locator(MultipleLocator(20))
# axt.yaxis.set_major_formatter(FormatStrFormatter('%d'))
# axt.yaxis.set_minor_locator(MultipleLocator(10))
# axt.xaxis.set_major_formatter(NullFormatter())
# axt.text(1.2, 0.5, r'$P\ (a)$', horizontalalignment='center', verticalalignment='center',
#             transform=axt.transAxes, rotation='vertical')
#-----------------------------------------------------------------------
# PARSING A STRING MATHEMATICAL EXPRESSION
# import parser
# import numpy
# t = 2
# st = parser.expr('t**3+4')
# code = st.compile()
# eval(code)
# t = numpy.arange(3)
# eval(code)
#-----------------------------------------------------------------------
# MATPLOTLIB DISCRETE COLOR BAR
# from matplotlib import cm
# cmap = cm.get_cmap('inferno', 6)

import os
import time
import numpy
import glob

from astropy.io import fits
import astropy.constants

#import matplotlib
#matplotlib.use('Agg')
from matplotlib import pyplot, rc, ticker, colors

from mangadap.dapfits import DAPQualityBitMask, DAPMapsBitMask
from mangadap.config.defaults import default_dap_method_path

#-----------------------------------------------------------------------------

def twod_bin(x, y, x0, dx, nx, y0, dy, ny, z=None):
    _z = numpy.ones(x.shape, dtype=float) if z is None else z

    bini = ((x-x0)/dx).astype(int)
    bini[ (bini < 0) | (bini >= nx) ] = -1
    binj = ((y-y0)/dy).astype(int)
    binj[ (binj < 0) | (binj >= ny) ] = -1

    n = numpy.zeros((nx,ny),dtype=int)
#    m = numpy.zeros((nx,ny),dtype=float)
#    d = numpy.zeros((nx,ny),dtype=float)
    for i in range(nx):
        iindx = bini == i
        for j in range(ny):
            jindx = iindx & (binj == j)
            if numpy.sum(jindx) == 0:
                continue
            n[i,j] = numpy.sum(jindx)
#            m[i,j] = numpy.median(_z[jindx])
#            d[i,j] = median_absolute_deviation(_z[jindx])

    return n if z is None else (n,m,d)


def d4000_sigma_haew_dist(fits_file, n, d4000_lim, sigma_lim, haew_lim, min_sn=None):

    # Channels with the H-alpha data and the D4000 data
    hac = 18
    d4c = 43

    # Set the method and get the root directory
    method = 'SPX-GAU-MILESHC'
    odir = default_dap_method_path(method)
    print('Output directory: {0}'.format(odir))

    # Find the intermediate files with the stellar-continuum models
    srchstr = os.path.join(odir, '*', '*', 'manga-*-MAPS-{0}.fits.gz'.format(method))
    print('Search string: {0}'.format(srchstr))
    maps_files = glob.glob(srchstr)
    nfiles = len(maps_files)
    print('Found {0} MAPS files.'.format(nfiles))

    d4000_bins = numpy.linspace(*d4000_lim,n+1)
    sigma_bins = numpy.logspace(*numpy.log10(sigma_lim),n+1)
    haew_bins = numpy.logspace(*numpy.log10(haew_lim),n+1)

    dl = d4000_bins[0]
    dd = d4000_bins[1]- d4000_bins[0]
    sl = numpy.log10(sigma_bins[0])
    ds = numpy.mean(numpy.diff(numpy.log10(sigma_bins)))
    hl = numpy.log10(haew_bins[0])
    dh = numpy.mean(numpy.diff(numpy.log10(haew_bins)))

    d4000_vs, vs_sigma = numpy.meshgrid(d4000_bins[:-1]+dd/2,
                                        sigma_bins[:-1]+ds*sigma_bins[:-1]*numpy.log(10)/2)
    d4000_vs, vs_haew = numpy.meshgrid(d4000_bins[:-1]+dd/2,
                                       haew_bins[:-1]+dh*haew_bins[:-1]*numpy.log(10)/2)
    sigma_vs, vs_haew = numpy.meshgrid(sigma_bins[:-1]+ds*sigma_bins[:-1]*numpy.log(10)/2,
                                        haew_bins[:-1]+ds*haew_bins[:-1]*numpy.log(10)/2)

    bin_d4000_vs_sigma = numpy.zeros((n,n), dtype=int)
    bin_d4000_vs_haew = numpy.zeros((n,n), dtype=int)
    bin_sigma_vs_haew = numpy.zeros((n,n), dtype=int)

    qual_bm = DAPQualityBitMask()
    maps_bm = DAPMapsBitMask()

    for fi in range(nfiles):
#    for fi in range(200):
        print('{0}/{1}'.format(fi+1,nfiles), end='\r')
        maps_file = maps_files[fi]
        print(maps_file)

        # Get the bin IDs, mapped to the correct spaxels
        hdu = fits.open(maps_file)

        # Check if the file was CRITICAL
        if qual_bm.flagged(hdu['PRIMARY'].header['DAPQUAL'], flag='DRPCRIT'):
            print('DRPCRIT')
            print(maps_file)
            hdu.close()
            del hdu
            continue

        plt, ifu = tuple([ int(x) for x in hdu[0].header['PLATEIFU'].split('-')])
        nsa_z = hdu['PRIMARY'].header['SCINPVEL'] / astropy.constants.c.to('km/s').value
        binid_map = hdu['BINID'].data.copy()

        spatial_shape = binid_map.shape
        binid, bin_indx = numpy.unique(binid_map.ravel(), return_index=True)
        binid = binid[1:]
        bin_indx = bin_indx[1:]

        # Get the D4000, H-alpha EW, and sigma values
#        snr = hdu['SPX_SNR'].data.ravel()[bin_indx]
#        snr_indx = snr > 1
#        good_snr = snr[snr_indx]
#        binid_snr = binid[snr_indx]

        mask = (hdu['EMLINE_GEW'].data[hac,:,:].ravel()[bin_indx] > 0) \
               & (hdu['SPECINDEX'].data[d4c,:,:].ravel()[bin_indx] > 0) \
               & (hdu['STELLAR_SIGMA'].data.ravel()[bin_indx] > 0)
        if min_sn is not None:
            mask &= (hdu['SPX_SNR'].data.ravel()[bin_indx] > min_sn)

        haew = hdu['EMLINE_GEW'].data[hac,:,:].ravel()[bin_indx][mask] #[snr_indx]
        d4000 = hdu['SPECINDEX'].data[d4c,:,:].ravel()[bin_indx][mask] #[snr_indx]
        sigma = hdu['STELLAR_SIGMA'].data.ravel()[bin_indx][mask] #[snr_indx]

        hdu.close()
        del hdu

#        bin_d4000_vs_sigma += twod_bin(numpy.log10(d4000), numpy.log10(sigma),
        bin_d4000_vs_sigma += twod_bin(d4000, numpy.log10(sigma),
                                       dl, dd, n, sl, ds, n)
#        bin_d4000_vs_haew += twod_bin(numpy.log10(d4000), numpy.log10(haew),
        bin_d4000_vs_haew += twod_bin(d4000, numpy.log10(haew),
                                      dl, dd, n, hl, dh, n)
        bin_sigma_vs_haew += twod_bin(numpy.log10(sigma), numpy.log10(haew),
                                      sl, ds, n, hl, dh, n)

    fits.HDUList([ fits.PrimaryHDU(),
                   fits.ImageHDU(data=bin_d4000_vs_sigma, name='D4000SIGMA'),
                   fits.ImageHDU(data=bin_d4000_vs_haew, name='D4000HAEW'),
                   fits.ImageHDU(data=bin_sigma_vs_haew, name='SIGMAHAEW')
                  ]).writeto(fits_file, overwrite=True)


def plot_dist_3d(fits_file, n, d4000_lim, sigma_lim, haew_lim):

    # Set the bin limits (not read from file!!)
    d4000_bins = numpy.linspace(*d4000_lim,n+1)
    sigma_bins = numpy.logspace(*numpy.log10(sigma_lim),n+1)
    haew_bins = numpy.logspace(*numpy.log10(haew_lim),n+1)

    dl = d4000_bins[0]
    dd = d4000_bins[1]- d4000_bins[0]
    sl = numpy.log10(sigma_bins[0])
    ds = numpy.mean(numpy.diff(numpy.log10(sigma_bins)))
    hl = numpy.log10(haew_bins[0])
    dh = numpy.mean(numpy.diff(numpy.log10(haew_bins)))

    d4000_vs, vs_sigma = numpy.meshgrid(d4000_bins[:-1]+dd/2,
                                        sigma_bins[:-1]+ds*sigma_bins[:-1]*numpy.log(10)/2)
    d4000_vs, vs_haew = numpy.meshgrid(d4000_bins[:-1]+dd/2,
                                       haew_bins[:-1]+dh*haew_bins[:-1]*numpy.log(10)/2)
    sigma_vs, vs_haew = numpy.meshgrid(sigma_bins[:-1]+ds*sigma_bins[:-1]*numpy.log(10)/2,
                                        haew_bins[:-1]+ds*haew_bins[:-1]*numpy.log(10)/2)

    hdu = fits.open(fits_file)

    bin_d4000_vs_sigma = hdu['D4000SIGMA'].data.copy()
    bin_d4000_vs_haew = hdu['D4000HAEW'].data.copy()
    bin_sigma_vs_haew = hdu['SIGMAHAEW'].data.copy()

    font = { 'size' : 12 }
    rc('font', **font)

    n_lim = numpy.amax([ numpy.amax(bin_d4000_vs_sigma), numpy.amax(bin_d4000_vs_haew),
                  numpy.amax(bin_sigma_vs_haew) ])
    print(n_lim)
    levels = [n_lim/32, n_lim/16, n_lim/8, n_lim/4, n_lim/2]

    _bin_d4000_vs_sigma = numpy.ma.MaskedArray(bin_d4000_vs_sigma.T,
                                               mask=numpy.invert(bin_d4000_vs_sigma.T>0))
    _bin_d4000_vs_haew = numpy.ma.MaskedArray(bin_d4000_vs_haew.T,
                                              mask=numpy.invert(bin_d4000_vs_haew.T>0))
    _bin_sigma_vs_haew = numpy.ma.MaskedArray(bin_sigma_vs_haew.T,
                                               mask=numpy.invert(bin_sigma_vs_haew.T>0))
    
    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    cax = fig.add_axes([0.66, 0.66, 0.2, 0.02])

    ax = fig.add_axes([0.15, 0.56, 0.4, 0.4], facecolor='0.95')
    ax.minorticks_on()
    ax.set_xlim(d4000_lim)
    ax.set_ylim(sigma_lim)
    ax.set_yscale('log')
    ax.grid(True, which='major', color='0.85', zorder=1, linestyle='-')
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
#    ax.set_xscale('log')
#    img = ax.imshow(bin_d4000_vs_sigma.T, origin='lower', interpolation='nearest',
#              extent=extent_d4000_vs_sigma, aspect='auto', zorder=2, vmin=0, vmax=n_lim)
    img = ax.pcolormesh(d4000_bins, sigma_bins, _bin_d4000_vs_sigma,
                        zorder=3, norm=colors.LogNorm(vmin=1, vmax=n_lim), lw=0, rasterized=True)
    ax.contour(d4000_vs, vs_sigma, _bin_d4000_vs_sigma, zorder=4,
               levels=levels, linewidths=0.5, norm=colors.LogNorm(vmin=1/3, vmax=n_lim/3))
    pyplot.colorbar(img, cax=cax, orientation='horizontal') #, ticks=ticker.MultipleLocator(200))
    ax.text(-0.2, 0.5, r'$\sigma_{\rm obs}$ (km/s)', horizontalalignment='center',
            verticalalignment='center', transform=ax.transAxes, rotation='vertical')
    cax.text(0.5, -3, r'$N_{\rm spaxel}$', horizontalalignment='center',
             verticalalignment='center', transform=cax.transAxes)

    cax.text(0.5, 8, r'MPL-6', horizontalalignment='center',
             verticalalignment='center', transform=cax.transAxes, fontsize=20)
    cax.text(0.5, 6, r'S/N > 10', horizontalalignment='center',
             verticalalignment='center', transform=cax.transAxes, fontsize=20)

    ax = fig.add_axes([0.15, 0.15, 0.4, 0.4], facecolor='0.95')
    ax.minorticks_on()
    ax.set_xlim(d4000_lim)
    ax.set_ylim(haew_lim)
    ax.set_yscale('log')
#    ax.set_xscale('log')
#    ax.imshow(bin_d4000_vs_haew.T, origin='lower', interpolation='nearest',
#              extent=extent_d4000_vs_haew, aspect='auto', zorder=2, vmin=0, vmax=n_lim)
    img = ax.pcolormesh(d4000_bins, haew_bins, _bin_d4000_vs_haew,
                        zorder=3, norm=colors.LogNorm(vmin=1, vmax=n_lim), lw=0, rasterized=True)
    ax.contour(d4000_vs, vs_haew, _bin_d4000_vs_haew, zorder=4,
               levels=levels, linewidths=0.5, norm=colors.LogNorm(vmin=1/3, vmax=n_lim/3))
    ax.text(-0.2, 0.5, r'H$\alpha$ EW ($\AA$)', horizontalalignment='center',
            verticalalignment='center', transform=ax.transAxes, rotation='vertical')
    ax.text(0.5, -0.15, r'D4000', horizontalalignment='center',
            verticalalignment='center', transform=ax.transAxes)

    ax = fig.add_axes([0.56, 0.15, 0.4, 0.4], facecolor='0.95')
    ax.minorticks_on()
    ax.set_xlim(sigma_lim)
    ax.set_ylim(haew_lim)
    ax.set_yscale('log')
    ax.set_xscale('log')
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
#    ax.imshow(bin_d4000_vs_haew.T, origin='lower', interpolation='nearest',
#              extent=extent_sigma_vs_haew, aspect='auto', zorder=2, vmin=0, vmax=n_lim)
    img = ax.pcolormesh(sigma_bins, haew_bins, _bin_sigma_vs_haew,
                        zorder=3, norm=colors.LogNorm(vmin=1, vmax=n_lim), lw=0, rasterized=True)
    ax.contour(sigma_vs, vs_haew, _bin_sigma_vs_haew, zorder=4,
               levels=levels, linewidths=0.5, norm=colors.LogNorm(vmin=1/3, vmax=n_lim/3))
    ax.text(0.5, -0.15, r'$\sigma_{\rm obs}$ (km/s)', horizontalalignment='center',
            verticalalignment='center', transform=ax.transAxes)

    plot_file = 'manga_representative.pdf'
    plot_file = None
    if plot_file is None:
        pyplot.show()
    else:
        fig.canvas.print_figure(plot_file, bbox_inches='tight')
    fig.clear()
    pyplot.close(fig)
    

    print('DONE                     ')


# Include binning with S/N
def d4000_sigma_haew_snr_dist(fits_file, n, d4000_lim, sigma_lim, haew_lim, snr_lim):

    # Channels with the H-alpha data and the D4000 data
    hac = 18
    d4c = 43

    # Set the method and get the root directory
    method = 'SPX-GAU-MILESHC'
    odir = default_dap_method_path(method)
    print('Output directory: {0}'.format(odir))

    # Find the intermediate files with the stellar-continuum models
    srchstr = os.path.join(odir, '*', '*', 'manga-*-MAPS-{0}.fits.gz'.format(method))
    print('Search string: {0}'.format(srchstr))
    maps_files = glob.glob(srchstr)
    nfiles = len(maps_files)
    print('Found {0} MAPS files.'.format(nfiles))

    d4000_bins = numpy.linspace(*d4000_lim,n+1)
    sigma_bins = numpy.logspace(*numpy.log10(sigma_lim),n+1)
    haew_bins = numpy.logspace(*numpy.log10(haew_lim),n+1)
    snr_bins = numpy.logspace(*numpy.log10(snr_lim),n+1)

    dl = d4000_bins[0]
    dd = d4000_bins[1]- d4000_bins[0]
    sl = numpy.log10(sigma_bins[0])
    ds = numpy.mean(numpy.diff(numpy.log10(sigma_bins)))
    hl = numpy.log10(haew_bins[0])
    dh = numpy.mean(numpy.diff(numpy.log10(haew_bins)))
    nl = numpy.log10(snr_bins[0])
    dn = numpy.mean(numpy.diff(numpy.log10(snr_bins)))

    d4000_vs, vs_sigma = numpy.meshgrid(d4000_bins[:-1]+dd/2,
                                        sigma_bins[:-1]+ds*sigma_bins[:-1]*numpy.log(10)/2)
    d4000_vs, vs_haew = numpy.meshgrid(d4000_bins[:-1]+dd/2,
                                       haew_bins[:-1]+dh*haew_bins[:-1]*numpy.log(10)/2)
    d4000_vs, vs_snr = numpy.meshgrid(d4000_bins[:-1]+dd/2,
                                      snr_bins[:-1]+dn*snr_bins[:-1]*numpy.log(10)/2)
    sigma_vs, vs_haew = numpy.meshgrid(sigma_bins[:-1]+ds*sigma_bins[:-1]*numpy.log(10)/2,
                                        haew_bins[:-1]+ds*haew_bins[:-1]*numpy.log(10)/2)
    sigma_vs, vs_snr = numpy.meshgrid(sigma_bins[:-1]+ds*sigma_bins[:-1]*numpy.log(10)/2,
                                      snr_bins[:-1]+dn*snr_bins[:-1]*numpy.log(10)/2)
    haew_vs, vs_snr = numpy.meshgrid(haew_bins[:-1]+dh*haew_bins[:-1]*numpy.log(10)/2,
                                     snr_bins[:-1]+dn*snr_bins[:-1]*numpy.log(10)/2)

    bin_d4000_vs_sigma = numpy.zeros((n,n), dtype=int)
    bin_d4000_vs_haew = numpy.zeros((n,n), dtype=int)
    bin_d4000_vs_snr = numpy.zeros((n,n), dtype=int)
    bin_sigma_vs_haew = numpy.zeros((n,n), dtype=int)
    bin_sigma_vs_snr = numpy.zeros((n,n), dtype=int)
    bin_haew_vs_snr = numpy.zeros((n,n), dtype=int)

    qual_bm = DAPQualityBitMask()
    maps_bm = DAPMapsBitMask()

    for fi in range(nfiles):
#    for fi in range(20):
        print('{0}/{1}'.format(fi+1,nfiles), end='\r')
        maps_file = maps_files[fi]
        print(maps_file)

        # Get the bin IDs, mapped to the correct spaxels
        hdu = fits.open(maps_file)

        # Check if the file was CRITICAL
        if qual_bm.flagged(hdu['PRIMARY'].header['DAPQUAL'], flag='DRPCRIT'):
            print('DRPCRIT')
            print(maps_file)
            hdu.close()
            del hdu
            continue

        plt, ifu = tuple([ int(x) for x in hdu[0].header['PLATEIFU'].split('-')])
        nsa_z = hdu['PRIMARY'].header['SCINPVEL'] / astropy.constants.c.to('km/s').value
        binid_map = hdu['BINID'].data[0,:,:].copy()

        spatial_shape = binid_map.shape
        binid, bin_indx = numpy.unique(binid_map.ravel(), return_index=True)
        binid = binid[1:]
        bin_indx = bin_indx[1:]

        # Get the D4000, H-alpha EW, and sigma values
#        snr = hdu['SPX_SNR'].data.ravel()[bin_indx]
#        snr_indx = snr > 1
#        good_snr = snr[snr_indx]
#        binid_snr = binid[snr_indx]

        mask = (hdu['EMLINE_GEW'].data[hac,:,:].ravel()[bin_indx] > 0) \
               & (hdu['SPECINDEX'].data[d4c,:,:].ravel()[bin_indx] > 0) \
               & (hdu['STELLAR_SIGMA'].data.ravel()[bin_indx] > 0) \
               & (hdu['SPX_SNR'].data.ravel()[bin_indx] > 0)

        haew = hdu['EMLINE_GEW'].data[hac,:,:].ravel()[bin_indx][mask]
        d4000 = hdu['SPECINDEX'].data[d4c,:,:].ravel()[bin_indx][mask]
        sigma = hdu['STELLAR_SIGMA'].data.ravel()[bin_indx][mask]
        snr = hdu['SPX_SNR'].data.ravel()[bin_indx][mask]

        hdu.close()
        del hdu

        bin_d4000_vs_sigma += twod_bin(d4000, numpy.log10(sigma), dl, dd, n, sl, ds, n)
        bin_d4000_vs_haew += twod_bin(d4000, numpy.log10(haew), dl, dd, n, hl, dh, n)
        bin_d4000_vs_snr += twod_bin(d4000, numpy.log10(snr), dl, dd, n, nl, dn, n)
        bin_sigma_vs_haew += twod_bin(numpy.log10(sigma), numpy.log10(haew), sl, ds, n, hl, dh, n)
        bin_sigma_vs_snr += twod_bin(numpy.log10(sigma), numpy.log10(snr), sl, ds, n, nl, dn, n)
        bin_haew_vs_snr += twod_bin(numpy.log10(haew), numpy.log10(snr), hl, dh, n, nl, dn, n)

    fits.HDUList([ fits.PrimaryHDU(),
                   fits.ImageHDU(data=bin_d4000_vs_sigma, name='D4000SIGMA'),
                   fits.ImageHDU(data=bin_d4000_vs_haew, name='D4000HAEW'),
                   fits.ImageHDU(data=bin_d4000_vs_snr, name='D4000SNR'),
                   fits.ImageHDU(data=bin_sigma_vs_haew, name='SIGMAHAEW'),
                   fits.ImageHDU(data=bin_sigma_vs_snr, name='SIGMASNR'),
                   fits.ImageHDU(data=bin_haew_vs_snr, name='HAEWSNR')
                  ]).writeto(fits_file, overwrite=True)


def plot_dist_4d(fits_file, n, d4000_lim, sigma_lim, haew_lim, snr_lim):

    # Set the bin limits (not read from file!!)
    d4000_bins = numpy.linspace(*d4000_lim,n+1)
    sigma_bins = numpy.logspace(*numpy.log10(sigma_lim),n+1)
    haew_bins = numpy.logspace(*numpy.log10(haew_lim),n+1)
    snr_bins = numpy.logspace(*numpy.log10(snr_lim),n+1)

    dl = d4000_bins[0]
    dd = d4000_bins[1]- d4000_bins[0]
    sl = numpy.log10(sigma_bins[0])
    ds = numpy.mean(numpy.diff(numpy.log10(sigma_bins)))
    hl = numpy.log10(haew_bins[0])
    dh = numpy.mean(numpy.diff(numpy.log10(haew_bins)))
    nl = numpy.log10(snr_bins[0])
    dn = numpy.mean(numpy.diff(numpy.log10(snr_bins)))

    d4000_vs, vs_sigma = numpy.meshgrid(d4000_bins[:-1]+dd/2,
                                        sigma_bins[:-1]+ds*sigma_bins[:-1]*numpy.log(10)/2)
    d4000_vs, vs_haew = numpy.meshgrid(d4000_bins[:-1]+dd/2,
                                       haew_bins[:-1]+dh*haew_bins[:-1]*numpy.log(10)/2)
    d4000_vs, vs_snr = numpy.meshgrid(d4000_bins[:-1]+dd/2,
                                      snr_bins[:-1]+dn*snr_bins[:-1]*numpy.log(10)/2)
    sigma_vs, vs_haew = numpy.meshgrid(sigma_bins[:-1]+ds*sigma_bins[:-1]*numpy.log(10)/2,
                                        haew_bins[:-1]+ds*haew_bins[:-1]*numpy.log(10)/2)
    sigma_vs, vs_snr = numpy.meshgrid(sigma_bins[:-1]+ds*sigma_bins[:-1]*numpy.log(10)/2,
                                      snr_bins[:-1]+dn*snr_bins[:-1]*numpy.log(10)/2)
    haew_vs, vs_snr = numpy.meshgrid(haew_bins[:-1]+dh*haew_bins[:-1]*numpy.log(10)/2,
                                     snr_bins[:-1]+dn*snr_bins[:-1]*numpy.log(10)/2)

    hdu = fits.open(fits_file)

    bin_d4000_vs_sigma = hdu['D4000SIGMA'].data.copy()
    bin_d4000_vs_haew = hdu['D4000HAEW'].data.copy()
    bin_d4000_vs_snr = hdu['D4000SNR'].data.copy()
    bin_sigma_vs_haew = hdu['SIGMAHAEW'].data.copy()
    bin_sigma_vs_snr = hdu['SIGMASNR'].data.copy()
    bin_haew_vs_snr = hdu['HAEWSNR'].data.copy()

    n_lim = numpy.amax([ numpy.amax(bin_d4000_vs_sigma),
                         numpy.amax(bin_d4000_vs_haew),
                         numpy.amax(bin_d4000_vs_snr),
                         numpy.amax(bin_sigma_vs_haew),
                         numpy.amax(bin_sigma_vs_snr),
                         numpy.amax(bin_haew_vs_snr) ])
    print(n_lim)
    levels = [n_lim/32, n_lim/16, n_lim/8, n_lim/4, n_lim/2]

    _bin_d4000_vs_sigma = numpy.ma.MaskedArray(bin_d4000_vs_sigma.T,
                                               mask=numpy.invert(bin_d4000_vs_sigma.T>0))
    _bin_d4000_vs_haew = numpy.ma.MaskedArray(bin_d4000_vs_haew.T,
                                              mask=numpy.invert(bin_d4000_vs_haew.T>0))
    _bin_d4000_vs_snr = numpy.ma.MaskedArray(bin_d4000_vs_snr.T,
                                             mask=numpy.invert(bin_d4000_vs_snr.T>0))
    _bin_sigma_vs_haew = numpy.ma.MaskedArray(bin_sigma_vs_haew.T,
                                               mask=numpy.invert(bin_sigma_vs_haew.T>0))
    _bin_sigma_vs_snr = numpy.ma.MaskedArray(bin_sigma_vs_snr.T,
                                             mask=numpy.invert(bin_sigma_vs_snr.T>0))
    _bin_haew_vs_snr = numpy.ma.MaskedArray(bin_haew_vs_snr.T,
                                             mask=numpy.invert(bin_haew_vs_snr.T>0))
    
    font = { 'size' : 12 }
    rc('font', **font)

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    cax = fig.add_axes([0.59, 0.79, 0.2, 0.02])

    ax = fig.add_axes([0.11, 0.68, 0.28, 0.28], facecolor='0.95')
    ax.minorticks_on()
    ax.set_xlim(d4000_lim)
    ax.set_ylim(sigma_lim)
    ax.set_yscale('log')
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    img = ax.pcolormesh(d4000_bins, sigma_bins, _bin_d4000_vs_sigma,
                        zorder=3, norm=colors.LogNorm(vmin=1, vmax=n_lim), lw=0, rasterized=True)
    ax.contour(d4000_vs, vs_sigma, _bin_d4000_vs_sigma, zorder=4,
               levels=levels, linewidths=0.5, norm=colors.LogNorm(vmin=1/3, vmax=n_lim/3))
    pyplot.colorbar(img, cax=cax, orientation='horizontal') #, ticks=ticker.MultipleLocator(200))
    ax.text(-0.25, 0.5, r'$\sigma_{\rm obs}$ (km/s)', horizontalalignment='center',
            verticalalignment='center', transform=ax.transAxes, rotation='vertical')
    cax.text(0.5, -3, r'$N_{\rm spaxel}$', horizontalalignment='center',
             verticalalignment='center', transform=cax.transAxes)

    cax.text(0.5, 4, r'MPL-6', horizontalalignment='center',
             verticalalignment='center', transform=cax.transAxes, fontsize=20)

    ax = fig.add_axes([0.11, 0.39, 0.28, 0.28], facecolor='0.95')
    ax.minorticks_on()
    ax.set_xlim(d4000_lim)
    ax.set_ylim(haew_lim)
    ax.set_yscale('log')
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    img = ax.pcolormesh(d4000_bins, haew_bins, _bin_d4000_vs_haew,
                        zorder=3, norm=colors.LogNorm(vmin=1, vmax=n_lim), lw=0, rasterized=True)
    ax.contour(d4000_vs, vs_haew, _bin_d4000_vs_haew, zorder=4,
               levels=levels, linewidths=0.5, norm=colors.LogNorm(vmin=1/3, vmax=n_lim/3))
    ax.text(-0.25, 0.5, r'H$\alpha$ EW ($\AA$)', horizontalalignment='center',
            verticalalignment='center', transform=ax.transAxes, rotation='vertical')

    ax = fig.add_axes([0.11, 0.10, 0.28, 0.28], facecolor='0.95')
    ax.minorticks_on()
    ax.set_xlim(d4000_lim)
    ax.set_ylim(snr_lim)
    ax.set_yscale('log')
    img = ax.pcolormesh(d4000_bins, snr_bins, _bin_d4000_vs_snr,
                        zorder=3, norm=colors.LogNorm(vmin=1, vmax=n_lim), lw=0, rasterized=True)
    ax.contour(d4000_vs, vs_snr, _bin_d4000_vs_snr, zorder=4,
               levels=levels, linewidths=0.5, norm=colors.LogNorm(vmin=1/3, vmax=n_lim/3))
    ax.text(-0.25, 0.5, r'S/N$_g$', horizontalalignment='center',
            verticalalignment='center', transform=ax.transAxes, rotation='vertical')
    ax.text(0.5, -0.2, r'D4000', horizontalalignment='center',
            verticalalignment='center', transform=ax.transAxes)

    ax = fig.add_axes([0.40, 0.39, 0.28, 0.28], facecolor='0.95')
    ax.minorticks_on()
    ax.set_xlim(sigma_lim)
    ax.set_ylim(haew_lim)
    ax.set_yscale('log')
    ax.set_xscale('log')
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    img = ax.pcolormesh(sigma_bins, haew_bins, _bin_sigma_vs_haew,
                        zorder=3, norm=colors.LogNorm(vmin=1, vmax=n_lim), lw=0, rasterized=True)
    ax.contour(sigma_vs, vs_haew, _bin_sigma_vs_haew, zorder=4,
               levels=levels, linewidths=0.5, norm=colors.LogNorm(vmin=1/3, vmax=n_lim/3))

    ax = fig.add_axes([0.40, 0.10, 0.28, 0.28], facecolor='0.95')
    ax.minorticks_on()
    ax.set_xlim(sigma_lim)
    ax.set_ylim(snr_lim)
    ax.set_yscale('log')
    ax.set_xscale('log')
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    img = ax.pcolormesh(sigma_bins, snr_bins, _bin_sigma_vs_snr,
                        zorder=3, norm=colors.LogNorm(vmin=1, vmax=n_lim), lw=0, rasterized=True)
    ax.contour(sigma_vs, vs_snr, _bin_sigma_vs_snr, zorder=4,
               levels=levels, linewidths=0.5, norm=colors.LogNorm(vmin=1/3, vmax=n_lim/3))
    ax.text(0.5, -0.20, r'$\sigma_{\rm obs}$ (km/s)', horizontalalignment='center',
            verticalalignment='center', transform=ax.transAxes)

    ax = fig.add_axes([0.69, 0.10, 0.28, 0.28], facecolor='0.95')
    ax.minorticks_on()
    ax.set_xlim(haew_lim)
    ax.set_ylim(snr_lim)
    ax.set_yscale('log')
    ax.set_xscale('log')
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    img = ax.pcolormesh(haew_bins, snr_bins, _bin_haew_vs_snr,
                        zorder=3, norm=colors.LogNorm(vmin=1, vmax=n_lim), lw=0, rasterized=True)
    ax.contour(haew_vs, vs_snr, _bin_haew_vs_snr, zorder=4,
               levels=levels, linewidths=0.5, norm=colors.LogNorm(vmin=1/3, vmax=n_lim/3))
    ax.text(0.5, -0.20, r'H$\alpha$ EW ($\AA$)', horizontalalignment='center',
            verticalalignment='center', transform=ax.transAxes)

    plot_file = 'manga_representative_4d.pdf'
#    plot_file = None
    if plot_file is None:
        pyplot.show()
    else:
        fig.canvas.print_figure(plot_file, bbox_inches='tight')
    fig.clear()
    pyplot.close(fig)
    

    print('DONE                     ')


#    fits.HDUList([ fits.PrimaryHDU(),
#                   fits.BinTableHDU.from_columns([ fits.Column(name=n,
#                                                               format=rec_to_fits_type(tbl[n]),
#                                                               array=tbl[n])
#                                                       for n in tbl.dtype.names ],
#                                                  name='PAR')
#                  ]).writeto('crap.fits.gz', overwrite=True)
#-----------------------------------------------------------------------------

if __name__ == '__main__':
    t = time.clock()

    # No S/N limit
    n=100
    d4000_lim=[0.7,3.5]
    sigma_lim=[20,600]
    haew_lim=[0.03,200]
    snr_lim=[1,200]

    fits_file='SPX-bin_snr.fits.gz'
#    overwrite=True
    overwrite=False
    if not os.path.isfile(fits_file) or overwrite:
        d4000_sigma_haew_snr_dist(fits_file, n, d4000_lim, sigma_lim, haew_lim, snr_lim)
    plot_dist_4d(fits_file, n, d4000_lim, sigma_lim, haew_lim, snr_lim)

    exit()

    # No S/N limit
    n=100
    d4000_lim=[0.7,3.5]
    sigma_lim=[20,600]
    haew_lim=[0.03,200]

    fits_file='SPX-snr_gt_00.fits.gz'
#    overwrite=True
    overwrite=False
    if not os.path.isfile(fits_file) or overwrite:
        d4000_sigma_haew_dist(fits_file, n, d4000_lim, sigma_lim, haew_lim)
    plot_dist_3d(fits_file, n, d4000_lim, sigma_lim, haew_lim)

    # Set S/N > 10 limit, but keep the same volume dimensions
    n=100
    d4000_lim=[0.7,3.5]
    sigma_lim=[20,600]
    haew_lim=[0.03,200]

    fits_file='SPX-snr_gt_10.fits.gz'
#    overwrite=True
    overwrite=False
    if not os.path.isfile(fits_file) or overwrite:
        d4000_sigma_haew_dist(fits_file, n, d4000_lim, sigma_lim, haew_lim, min_sn=10)
    plot_dist_3d(fits_file, n, d4000_lim, sigma_lim, haew_lim)

    # Repeat S/N > 10 limit, but with smaller d4000 dimension
    n=100
    d4000_lim=[0.8,2.6]
    sigma_lim=[20,600]
    haew_lim=[0.03,200]

    fits_file='SPX-snr_gt_10_v2.fits.gz'
#    overwrite=True
    overwrite=False
    if not os.path.isfile(fits_file) or overwrite:
        d4000_sigma_haew_dist(fits_file, n, d4000_lim, sigma_lim, haew_lim, mins_sn=10)
    plot_dist_3d(fits_file, n, d4000_lim, sigma_lim, haew_lim)
    print('Elapsed time: {0} seconds'.format(time.clock() - t))


